@@ -1,7 +1,11 @@
 #include <config.h>
 #include <libintl.h>
 #include <gtk/gtk.h>
+#if defined (GDK_WINDOWING_X11)
 #include <gdk/gdkx.h>
+#elif defined (GDK_WINDOWING_WIN32)
+#include <gdk/gdkwin32.h>
+#endif
 #include <gdk/gdkkeysyms.h>
 #include "eggcellrendererkeys.h"
 #include "eggaccelerators.h"
@@ -18,6 +22,8 @@
 
 #define EGG_CELL_RENDERER_TEXT_PATH "egg-cell-renderer-text"
 
+#define TOOLTIP_TEXT _("New accelerator...")
+
 static void             egg_cell_renderer_keys_finalize      (GObject             *object);
 static void             egg_cell_renderer_keys_init          (EggCellRendererKeys *cell_keys);
 static void             egg_cell_renderer_keys_class_init    (EggCellRendererKeysClass *cell_keys_class);
@@ -52,6 +58,7 @@ enum {
 
   PROP_ACCEL_KEY,
   PROP_ACCEL_MASK,
+  PROP_KEYCODE,
   PROP_ACCEL_MODE
 };
 
@@ -172,22 +179,32 @@ egg_cell_renderer_keys_class_init (EggCellRendererKeysClass *cell_keys_class)
                                                        GDK_TYPE_MODIFIER_TYPE,
                                                        0,
                                                        G_PARAM_READABLE | G_PARAM_WRITABLE));
+
+  g_object_class_install_property (object_class,
+		  		   PROP_KEYCODE,
+				   g_param_spec_uint ("keycode",
+					   	      _("Accelerator keycode"),
+						      _("Accelerator keycode"),
+						      0,
+						      G_MAXINT,
+						      0,
+						      G_PARAM_READABLE | G_PARAM_WRITABLE));
   
   /* FIXME: Register the enum when moving to GTK+ */
   g_object_class_install_property (object_class,
                                    PROP_ACCEL_MODE,
                                    g_param_spec_int ("accel_mode",
-						     _("Accelerator Mode"),
+						     _("Accel Mode"),
 						     _("The type of accelerator."),
 						     0,
 						     2,
 						     0,
 						     G_PARAM_READABLE | G_PARAM_WRITABLE));
   
-  g_signal_new ("keys_edited",
+  g_signal_new ("accel_edited",
                 EGG_TYPE_CELL_RENDERER_KEYS,
                 G_SIGNAL_RUN_LAST,
-                G_STRUCT_OFFSET (EggCellRendererKeysClass, keys_edited),
+                G_STRUCT_OFFSET (EggCellRendererKeysClass, accel_edited),
                 NULL, NULL,
                 marshal_VOID__STRING_UINT_FLAGS_UINT,
                 G_TYPE_NONE, 4,
@@ -195,6 +212,15 @@ egg_cell_renderer_keys_class_init (EggCellRendererKeysClass *cell_keys_class)
                 G_TYPE_UINT,
                 GDK_TYPE_MODIFIER_TYPE,
 		G_TYPE_UINT);
+
+  g_signal_new ("accel_cleared",
+                EGG_TYPE_CELL_RENDERER_KEYS,
+                G_SIGNAL_RUN_LAST,
+                G_STRUCT_OFFSET (EggCellRendererKeysClass, accel_cleared),
+                NULL, NULL,
+                g_cclosure_marshal_VOID__STRING,
+                G_TYPE_NONE, 1,
+		G_TYPE_STRING);
 }
 
 
@@ -213,12 +239,13 @@ egg_cell_renderer_keys_finalize (GObject *object)
 
 static gchar *
 convert_keysym_state_to_string (guint                  keysym,
+				guint		       keycode,
                                 EggVirtualModifierType mask)
 {
-  if (keysym == 0)
+  if (keysym == 0 && keycode == 0)
     return g_strdup (_("Disabled"));
   else
-    return egg_virtual_accelerator_name (keysym, mask);
+    return egg_virtual_accelerator_name (keysym, keycode, mask);
 }
 
 static void
@@ -269,14 +296,22 @@ egg_cell_renderer_keys_set_property  (GObject                  *object,
     case PROP_ACCEL_KEY:
       egg_cell_renderer_keys_set_accelerator (keys,
                                               g_value_get_uint (value),
+					      keys->keycode,
                                               keys->accel_mask);
       break;
 
     case PROP_ACCEL_MASK:
       egg_cell_renderer_keys_set_accelerator (keys,
                                               keys->accel_key,
+					      keys->keycode,
                                               g_value_get_flags (value));
       break;
+    case PROP_KEYCODE:
+      egg_cell_renderer_keys_set_accelerator (keys,
+		      			      keys->accel_key,
+					      g_value_get_uint (value),
+					      keys->accel_mask);
+      break;
 
     case PROP_ACCEL_MODE:
       egg_cell_renderer_keys_set_accel_mode (keys, g_value_get_int (value));
@@ -290,6 +325,7 @@ egg_cell_renderer_keys_set_property  (GObject                  *object,
 static gboolean 
 is_modifier (guint keycode)
 {
+#ifdef GDK_WINDOWING_X11
   gint i;
   gint map_size;
   XModifierKeymap *mod_keymap;
@@ -312,9 +348,23 @@ is_modifier (guint keycode)
   XFreeModifiermap (mod_keymap);
 
   return retval;
+#else
+  /* Bogus, doesn't take into account the fact that CapsLock is a
+   * modifier on some keyboard layouts. (I think?) 
+   */
+  return ((keycode == VK_SHIFT) ||
+	  (keycode == VK_LSHIFT) ||
+	  (keycode == VK_RSHIFT) ||
+	  (keycode == VK_CONTROL) ||
+	  (keycode == VK_LCONTROL) ||
+	  (keycode == VK_RCONTROL) ||
+	  (keycode == VK_MENU) ||
+	  (keycode == VK_LMENU) ||
+	  (keycode == VK_RMENU));
+#endif
 }
 
-void
+static void
 egg_cell_renderer_keys_get_size (GtkCellRenderer *cell,
 				 GtkWidget       *widget,
 				 GdkRectangle    *cell_area,
@@ -322,12 +372,13 @@ egg_cell_renderer_keys_get_size (GtkCellRenderer *cell,
 				 gint            *y_offset,
 				 gint            *width,
 				 gint            *height)
+
 {
   EggCellRendererKeys *keys = (EggCellRendererKeys *) cell;
   GtkRequisition requisition;
 
   if (keys->sizing_label == NULL)
-    keys->sizing_label = gtk_label_new (_("Type a new accelerator, or press Backspace to clear"));
+    keys->sizing_label = gtk_label_new (TOOLTIP_TEXT);
 
   gtk_widget_size_request (keys->sizing_label, &requisition);
   (* GTK_CELL_RENDERER_CLASS (parent_class)->get_size) (cell, widget, cell_area, x_offset, y_offset, width, height);
@@ -347,22 +398,23 @@ grab_key_callback (GtkWidget    *widget,
                    GdkEventKey  *event,
                    void         *data)
 {
-  GdkModifierType accel_mods;
+  GdkModifierType accel_mods = 0;
   guint accel_keyval;
   EggCellRendererKeys *keys;
   char *path;
   gboolean edited;
+  gboolean cleared;
   GdkModifierType consumed_modifiers;  
   guint upper;
   GdkModifierType ignored_modifiers;
   
   keys = EGG_CELL_RENDERER_KEYS (data);
-  accel_mods = 0;
 
   if (is_modifier (event->hardware_keycode))
     return TRUE;
 
   edited = FALSE;
+  cleared = FALSE;
 
   consumed_modifiers = 0;
   gdk_keymap_translate_keyboard_state (gdk_keymap_get_default (),
@@ -391,12 +443,20 @@ grab_key_callback (GtkWidget    *widget,
                                         EGG_VIRTUAL_SCROLL_LOCK_MASK,
                                         &ignored_modifiers);
   
+  /* http://bugzilla.gnome.org/show_bug.cgi?id=139605
+   * mouse keys should effect keybindings */
+  ignored_modifiers |=	GDK_BUTTON1_MASK |
+			GDK_BUTTON2_MASK |
+			GDK_BUTTON3_MASK |
+			GDK_BUTTON4_MASK |
+			GDK_BUTTON5_MASK;
+
   /* filter consumed/ignored modifiers */
 
   if (keys->accel_mode == EGG_CELL_RENDERER_KEYS_MODE_GTK)
-    accel_mods = event->state & ~(consumed_modifiers | ignored_modifiers);
+    accel_mods = event->state & GDK_MODIFIER_MASK & ~(consumed_modifiers | ignored_modifiers);
   else if (keys->accel_mode == EGG_CELL_RENDERER_KEYS_MODE_X)
-    accel_mods = event->state & ~(ignored_modifiers);
+    accel_mods = event->state & GDK_MODIFIER_MASK & ~(ignored_modifiers);
   else
     g_assert_not_reached ();
     
@@ -404,10 +464,11 @@ grab_key_callback (GtkWidget    *widget,
     goto out; /* cancel */
 
   /* clear the accelerator on Backspace */
-  if (keys->edit_key != 0 &&
-      accel_mods == 0 &&
-      accel_keyval == GDK_BackSpace)
-    accel_keyval = 0;
+  if (accel_mods == 0 && accel_keyval == GDK_BackSpace)
+    {
+      cleared = TRUE;
+      goto out;
+    }
 
   if (keys->accel_mode == EGG_CELL_RENDERER_KEYS_MODE_GTK)
     {
@@ -416,30 +477,31 @@ grab_key_callback (GtkWidget    *widget,
 	  accel_keyval = 0;
 	  accel_mods = 0;
 	}
-
-      /* Remove modifiers like super and hyper, as GTK+ ignores them. */
-      accel_mods = accel_mods & GDK_MODIFIER_MASK;
     }
   
   edited = TRUE;
  out:
-  path = g_strdup (g_object_get_data (G_OBJECT (keys->edit_widget),
-                                      EGG_CELL_RENDERER_TEXT_PATH));
-  
   gdk_keyboard_ungrab (event->time);
   gdk_pointer_ungrab (event->time);
   
+  path = g_strdup (g_object_get_data (G_OBJECT (keys->edit_widget), EGG_CELL_RENDERER_TEXT_PATH));
+
   gtk_cell_editable_editing_done (GTK_CELL_EDITABLE (keys->edit_widget));
   gtk_cell_editable_remove_widget (GTK_CELL_EDITABLE (keys->edit_widget));
   keys->edit_widget = NULL;
   keys->grab_widget = NULL;
   
   if (edited)
-    g_signal_emit_by_name (G_OBJECT (keys), "keys_edited", path,
-                           accel_keyval, accel_mods, event->hardware_keycode);
+    {
+      g_signal_emit_by_name (G_OBJECT (keys), "accel_edited", path,
+			     accel_keyval, accel_mods, event->hardware_keycode);
+    }
+  else if (cleared)
+    {
+      g_signal_emit_by_name (G_OBJECT (keys), "accel_cleared", path);
+    }
 
   g_free (path);
-  
   return TRUE;
 }
 
@@ -559,12 +621,8 @@ egg_cell_renderer_keys_start_editing (GtkCellRenderer      *cell,
   gtk_widget_modify_fg (label, GTK_STATE_NORMAL,
                         &widget->style->fg[GTK_STATE_SELECTED]);
   
-  if (keys->accel_key != 0)
-    gtk_label_set_text (GTK_LABEL (label),
-			_("Type a new accelerator, or press Backspace to clear"));
-  else
-    gtk_label_set_text (GTK_LABEL (label),
-                        _("Type a new accelerator"));
+  gtk_label_set_text (GTK_LABEL (label),
+		  TOOLTIP_TEXT);
 
   gtk_container_add (GTK_CONTAINER (eventbox), label);
   
@@ -582,9 +640,10 @@ egg_cell_renderer_keys_start_editing (GtkCellRenderer      *cell,
 }
 
 void
-egg_cell_renderer_keys_set_accelerator (EggCellRendererKeys    *keys,
-                                        guint                   keyval,
-                                        EggVirtualModifierType  mask)
+egg_cell_renderer_keys_set_accelerator (EggCellRendererKeys *keys,
+                                        guint                keyval,
+					guint		     keycode,
+					EggVirtualModifierType  mask)
 {
   char *text;
   gboolean changed;
@@ -610,13 +669,21 @@ egg_cell_renderer_keys_set_accelerator (EggCellRendererKeys    *keys,
       g_object_notify (G_OBJECT (keys), "accel_mask");
       changed = TRUE;
     }  
+
+  if (keycode != keys->keycode)
+    {
+      keys->keycode = keycode;
+
+      g_object_notify (G_OBJECT (keys), "keycode");
+      changed = TRUE;
+    }
   g_object_thaw_notify (G_OBJECT (keys));
 
   if (changed)
     {
       /* sync string to the key values */
       celltext = GTK_CELL_RENDERER_TEXT (keys);
-      text = convert_keysym_state_to_string (keys->accel_key, keys->accel_mask);
+      text = convert_keysym_state_to_string (keys->accel_key, keys->keycode, keys->accel_mask);
       g_object_set (keys, "text", text, NULL);
       g_free (text);
     }
